"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.options = exports.chainKotlinGeneratorContext = exports.packageToPath = void 0;
const generator_common_1 = require("@openapi-generator-plus/generator-common");
const handlebars_templates_1 = require("@openapi-generator-plus/handlebars-templates");
const java_like_generator_helper_1 = require("@openapi-generator-plus/java-like-generator-helper");
const types_1 = require("@openapi-generator-plus/types");
const handlebars_1 = __importDefault(require("handlebars"));
const path_1 = __importDefault(require("path"));
const fs_1 = require("fs");
function escapeString(value) {
    if (typeof value !== 'string' && typeof value !== 'number' && typeof value !== 'boolean') {
        throw new Error(`escapeString called with unsupported type: ${typeof value} (${value})`);
    }
    value = String(value);
    value = value.replace(/\\/g, '\\\\');
    value = value.replace(/"/g, '\\"');
    value = value.replace(/\r/g, '\\r');
    value = value.replace(/\n/g, '\\n');
    return value;
}
/**
 * Turns a Java package name into a path
 * @param packageName Java package name
 */
function packageToPath(packageName) {
    return packageName.replace(/\./g, path_1.default.sep);
}
exports.packageToPath = packageToPath;
function computeCustomTemplatesPath(configPath, customTemplatesPath) {
    if (configPath) {
        return path_1.default.resolve(path_1.default.dirname(configPath), customTemplatesPath);
    }
    else {
        return customTemplatesPath;
    }
}
function computeRelativeSourceOutputPath(config) {
    const gradle = config.gradle;
    const defaultPath = gradle ? path_1.default.join('src', 'main', 'kotlin') : '';
    return (0, generator_common_1.configString)(config, 'relativeSourceOutputPath', defaultPath);
}
function toSafeTypeForComposing(nativeType) {
    return nativeType;
}
function chainKotlinGeneratorContext(base, add) {
    const result = Object.assign(Object.assign({}, base), { loadAdditionalTemplates: async function (hbs) {
            /* Load the additional first, so that earlier contexts in the chain have priority */
            if (add.loadAdditionalTemplates) {
                await add.loadAdditionalTemplates(hbs);
            }
            if (base.loadAdditionalTemplates) {
                await base.loadAdditionalTemplates(hbs);
            }
        }, additionalWatchPaths: function () {
            const result = [];
            if (base.additionalWatchPaths) {
                result.push(...base.additionalWatchPaths());
            }
            if (add.additionalWatchPaths) {
                result.push(...add.additionalWatchPaths());
            }
            return result;
        }, additionalExportTemplates: async function (outputPath, doc, hbs, rootContext) {
            if (base.additionalExportTemplates) {
                await base.additionalExportTemplates(outputPath, doc, hbs, rootContext);
            }
            if (add.additionalExportTemplates) {
                await add.additionalExportTemplates(outputPath, doc, hbs, rootContext);
            }
        } });
    return result;
}
exports.chainKotlinGeneratorContext = chainKotlinGeneratorContext;
/* https://kotlinlang.org/docs/keyword-reference.html#hard-keywords */
const RESERVED_WORDS = [
    'as', 'break', 'class', 'continue', 'do', 'else', 'false', 'for', 'fun', 'if', 'in', 'interface', 'is', 'null', 'object', 'package',
    'return', 'super', 'this', 'throw', 'true', 'try', 'typealias', 'typeof', 'val', 'var', 'when', 'while',
    'unknown', // for our enum cases
];
function options(config, context) {
    const packageName = (0, generator_common_1.configString)(config, 'package', 'com.example');
    const apiPackage = (0, generator_common_1.configString)(config, 'apiPackage', packageName);
    const customTemplates = (0, generator_common_1.configString)(config, 'customTemplates', undefined);
    const gradle = (0, generator_common_1.configObject)(config, 'gradle', undefined);
    const options = Object.assign(Object.assign({}, (0, java_like_generator_helper_1.options)(config, createJavaLikeContext(context))), { apiPackage, modelPackage: (0, generator_common_1.configString)(config, 'modelPackage', `${packageName}.model`), supportPackage: (0, generator_common_1.configString)(config, 'supportPackage', `${packageName}.support`), securityPackage: (0, generator_common_1.configString)(config, 'securityPackage', `${packageName}.security`), customTemplatesPath: customTemplates ? computeCustomTemplatesPath(config.configPath, customTemplates) : null, hideGenerationTimestamp: (0, generator_common_1.configBoolean)(config, 'hideGenerationTimestamp', false), dateImplementation: (0, generator_common_1.configString)(config, 'dateImplementation', 'kotlinx.datetime.LocalDate'), timeImplementation: (0, generator_common_1.configString)(config, 'timeImplementation', 'kotlinx.datetime.LocalTime'), dateTimeImplementation: (0, generator_common_1.configString)(config, 'dateTimeImplementation', 'kotlinx.datetime.Instant'), binaryRepresentation: (0, generator_common_1.configString)(config, 'binaryRepresentation', 'kotlin.ByteArray'), gradle: gradle ? {
            groupId: (0, generator_common_1.configString)(gradle, 'groupId', 'com.example', 'gradle.'),
            version: (0, generator_common_1.configString)(gradle, 'version', '0.0.1', 'gradle.'),
            versions: (0, generator_common_1.configObject)(gradle, 'versions', {}, 'gradle.'),
        } : null, relativeSourceOutputPath: computeRelativeSourceOutputPath(config) });
    return options;
}
exports.options = options;
function createJavaLikeContext(context) {
    const javaLikeContext = Object.assign(Object.assign({}, context), { reservedWords: () => RESERVED_WORDS, defaultConstantStyle: "camelCase" /* ConstantStyle.camelCase */ });
    return javaLikeContext;
}
function createGenerator(config, context) {
    const generatorOptions = options(config, context);
    const baseGenerator = context.baseGenerator(config, context);
    const aCommonGenerator = (0, generator_common_1.commonGenerator)(config, context);
    return Object.assign(Object.assign(Object.assign(Object.assign({}, baseGenerator), aCommonGenerator), (0, java_like_generator_helper_1.javaLikeGenerator)(config, createJavaLikeContext(context))), { toLiteral: (value, options) => {
            if (value === undefined) {
                const defaultValue = context.generator().defaultValue(options);
                if (defaultValue === null) {
                    return null;
                }
                return defaultValue.literalValue;
            }
            const { type, format, nullable, schemaType } = options;
            if (value === null) {
                if (nullable) {
                    return 'null';
                }
                const defaultValue = context.generator().defaultValue(options);
                if (defaultValue === null) {
                    return null;
                }
                return defaultValue.literalValue;
            }
            if (schemaType === types_1.CodegenSchemaType.ENUM) {
                return `${options.nativeType.concreteType}.${context.generator().toEnumMemberName(String(value))}`;
            }
            switch (type) {
                case 'integer': {
                    if (typeof value === 'string') {
                        const parsedValue = parseInt(value);
                        if (isNaN(parsedValue)) {
                            throw new Error(`toLiteral with type integer called with non-number: ${typeof value} (${value})`);
                        }
                        value = parsedValue;
                    }
                    if (typeof value !== 'number') {
                        throw new Error(`toLiteral with type integer called with non-number: ${typeof value} (${value})`);
                    }
                    if (format === 'int32' || !format) {
                        return `${value}`;
                    }
                    else if (format === 'int64') {
                        return `${value}L`;
                    }
                    else {
                        throw new Error(`Unsupported ${type} format: ${format}`);
                    }
                }
                case 'number': {
                    if (typeof value === 'string') {
                        const parsedValue = parseFloat(value);
                        if (isNaN(parsedValue)) {
                            throw new Error(`toLiteral with type number called with non-number: ${typeof value} (${value})`);
                        }
                        value = parsedValue;
                    }
                    if (typeof value !== 'number') {
                        throw new Error(`toLiteral with type number called with non-number: ${typeof value} (${value})`);
                    }
                    if (format === 'float') {
                        return `${value}f`;
                    }
                    else if (format === 'double' || !format) {
                        return `${value}`;
                    }
                    else {
                        throw new Error(`Unsupported ${type} format: ${format}`);
                    }
                }
                case 'string': {
                    if (typeof value !== 'string' && typeof value !== 'number' && typeof value !== 'boolean') {
                        throw new Error(`toLiteral with type string called with unsupported type: ${typeof value} (${value})`);
                    }
                    if (format === 'byte') {
                        return `"${escapeString(value)}"`;
                    }
                    else if (format === 'binary') {
                        return `"${escapeString(value)}".toByteArray(kotlin.text.Charsets.UTF_8)`;
                    }
                    else if (format === 'date') {
                        return `${generatorOptions.dateImplementation}.parse("${escapeString(value)}")`;
                    }
                    else if (format === 'time') {
                        return `${generatorOptions.timeImplementation}.parse("${escapeString(value)}")`;
                    }
                    else if (format === 'date-time') {
                        return `${generatorOptions.dateTimeImplementation}.parse("${escapeString(value)}")`;
                    }
                    else if (format === 'uuid') {
                        return `java.util.UUID.fromString("${escapeString(value)}")`;
                    }
                    else {
                        return `"${escapeString(value)}"`;
                    }
                }
                case 'boolean':
                    if (typeof value === 'string') {
                        value = value.toLowerCase() === 'true';
                    }
                    if (typeof value === 'number') {
                        value = value !== 0;
                    }
                    if (typeof value !== 'boolean') {
                        throw new Error(`toLiteral with type boolean called with non-boolean: ${typeof value} (${value})`);
                    }
                    return `${value}`;
                case 'object':
                    if (typeof value === 'string') {
                        if (value) {
                            return value;
                        }
                        else {
                            return 'null';
                        }
                    }
                    else {
                        context.log(types_1.CodegenLogLevel.WARN, `Literal is unsupported for schema type object: ${(0, generator_common_1.debugStringify)(value)}`);
                        return 'null';
                    }
                case 'anyOf':
                case 'oneOf':
                    context.log(types_1.CodegenLogLevel.WARN, `Literal value of type ${typeof value} is unsupported for schema type object: ${(0, generator_common_1.debugStringify)(value)}`);
                    return 'null';
                case 'file':
                    throw new Error(`Cannot format literal for type ${type}`);
                case 'array': {
                    const arrayValue = Array.isArray(value) ? value : [value];
                    const component = options.component;
                    if (!component) {
                        throw new Error(`toLiteral cannot format array literal without a component type: ${value}`);
                    }
                    return `listOf(${arrayValue.map(v => context.generator().toLiteral(v, Object.assign(Object.assign({}, component.schema), component))).join(', ')})`;
                }
            }
            throw new Error(`Unsupported literal type name "${type}" in options: ${(0, generator_common_1.debugStringify)(options)}`);
        }, toNativeType: (options) => {
            const { format, schemaType, vendorExtensions } = options;
            /* Note that we return separate componentTypes in this function in case the type
               is transformed, using nativeTypeTransformer, and the native type becomes primitive
               as the component type must still be non-primitive.
             */
            if (vendorExtensions && vendorExtensions['x-kotlin-type']) {
                return new context.NativeType(String(vendorExtensions['x-kotlin-type']));
            }
            /* See https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#data-types */
            switch (schemaType) {
                case types_1.CodegenSchemaType.INTEGER: {
                    // TODO, DISCUSS WITH KARL & GRANT : Really...? Why do we want to use int which only has a precision of 32 bit when we are unsure of its format 
                    // instead of long which has more precision and thus support for larger numbers?
                    // 
                    // See also: 
                    // - https://swagger.io/docs/specification/data-models/data-types/#numbers
                    // - https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Number#number_encoding
                    // - https://www.w3schools.com/js/js_numbers.asp
                    if (format === 'int32' || !format) {
                        return new context.NativeType('kotlin.Int');
                    }
                    else if (format === 'int64') {
                        return new context.NativeType('kotlin.Long');
                    }
                    else {
                        throw new Error(`Unsupported integer format: ${format}`);
                    }
                }
                case types_1.CodegenSchemaType.NUMBER: {
                    if (format === 'float') {
                        return new context.NativeType('kotlin.Float');
                    }
                    else if (format === 'double' || !format) {
                        return new context.NativeType('kotlin.Double');
                    }
                    else {
                        throw new Error(`Unsupported number format: ${format}`);
                    }
                }
                case types_1.CodegenSchemaType.DATE:
                    return new context.NativeType(generatorOptions.dateImplementation, {
                        serializedType: 'String',
                    });
                case types_1.CodegenSchemaType.TIME:
                    return new context.NativeType(generatorOptions.timeImplementation, {
                        serializedType: 'String',
                    });
                case types_1.CodegenSchemaType.DATETIME:
                    return new context.NativeType(generatorOptions.dateTimeImplementation, {
                        serializedType: 'String',
                    });
                case types_1.CodegenSchemaType.STRING: {
                    if (format === 'byte') {
                        /* base64 encoded characters */
                        return new context.NativeType('String');
                    }
                    else if (format === 'uuid') {
                        return new context.NativeType('java.util.UUID', {
                            serializedType: 'String',
                        });
                    }
                    else if (format === 'url') {
                        return new context.NativeType('java.net.URL', {
                            serializedType: 'String',
                        });
                    }
                    else {
                        return new context.NativeType('kotlin.String');
                    }
                }
                case types_1.CodegenSchemaType.BOOLEAN: {
                    return new context.NativeType('kotlin.Boolean');
                }
                case types_1.CodegenSchemaType.BINARY: {
                    return new context.NativeType(generatorOptions.binaryRepresentation);
                }
            }
            throw new Error(`Unsupported schema type: ${schemaType}`);
        }, toNativeObjectType: function (options) {
            const { scopedName } = options;
            let modelName = `${generatorOptions.modelPackage}`;
            for (const name of scopedName) {
                modelName += `.${context.generator().toClassName(name)}`;
            }
            return new context.NativeType(modelName);
        }, toNativeArrayType: (options) => {
            const { componentNativeType, uniqueItems } = options;
            if (uniqueItems) {
                return new context.TransformingNativeType(componentNativeType, {
                    default: (nativeType) => `kotlin.collections.List<${(nativeType.componentType || nativeType).nativeType}>`,
                    literalType: () => 'kotlin.collections.List',
                    concreteType: (nativeType) => `${(nativeType.componentType || nativeType).nativeType}`,
                });
            }
            else {
                return new context.TransformingNativeType(componentNativeType, {
                    default: (nativeType) => `kotlin.collections.List<${(nativeType.componentType || nativeType).nativeType}>`,
                    literalType: () => 'kotlin.collections.List',
                    concreteType: (nativeType) => `${(nativeType.componentType || nativeType).nativeType}`,
                });
            }
        }, toNativeMapType: (options) => {
            const { keyNativeType, componentNativeType } = options;
            return new context.ComposingNativeType([keyNativeType, componentNativeType], {
                default: ([keyNativeType, componentNativeType]) => `kotlin.collections.Map<${(keyNativeType.componentType || keyNativeType).nativeType}, ${(componentNativeType.componentType || componentNativeType).nativeType}>`,
                literalType: () => 'kotlin.collections.Map',
                concreteType: ([keyNativeType, componentNativeType]) => `${(keyNativeType.componentType || keyNativeType).nativeType}, ${(componentNativeType.componentType || componentNativeType).nativeType}`,
            });
        }, nativeTypeUsageTransformer: ({ nullable, required }) => ({
            default: function (nativeType, nativeTypeString) {
                if (nullable) {
                    nativeTypeString = `${generatorOptions.supportPackage}.Nullable<${toSafeTypeForComposing(nativeTypeString)}>`;
                }
                if (!required) {
                    nativeTypeString = `${toSafeTypeForComposing(nativeTypeString)}?`;
                }
                return nativeTypeString;
            },
            /* We don't transform the concrete type as the concrete type is never null; we use it to make new objects */
            concreteType: null,
            parentType: null,
        }), toSuggestedSchemaName: (name, options) => {
            if (options.schemaType === types_1.CodegenSchemaType.ENUM) {
                name = `${name}`;
            }
            else if (options.purpose === types_1.CodegenSchemaPurpose.EXTRACTED_INTERFACE) {
                name = `i_${name}`;
            }
            else if (options.purpose === types_1.CodegenSchemaPurpose.ABSTRACT_IMPLEMENTATION) {
                name = `abstract_${name}`;
            }
            else if (options.purpose === types_1.CodegenSchemaPurpose.IMPLEMENTATION) {
                name = `${name}_impl`;
            }
            return name;
        }, defaultValue: (options) => {
            const { schemaType, required } = options;
            if (!required) {
                return { value: null, literalValue: 'null' };
            }
            switch (schemaType) {
                case types_1.CodegenSchemaType.ENUM:
                case types_1.CodegenSchemaType.DATE:
                case types_1.CodegenSchemaType.TIME:
                case types_1.CodegenSchemaType.DATETIME:
                case types_1.CodegenSchemaType.BINARY:
                case types_1.CodegenSchemaType.OBJECT:
                case types_1.CodegenSchemaType.STRING:
                case types_1.CodegenSchemaType.ARRAY:
                case types_1.CodegenSchemaType.MAP:
                case types_1.CodegenSchemaType.INTERFACE:
                    return { value: null, literalValue: 'null' };
                case types_1.CodegenSchemaType.NUMBER: {
                    const literalValue = context.generator().toLiteral(0.0, options);
                    if (literalValue === null) {
                        return null;
                    }
                    return { value: 0.0, literalValue };
                }
                case types_1.CodegenSchemaType.INTEGER: {
                    const literalValue = context.generator().toLiteral(0, options);
                    if (literalValue === null) {
                        return null;
                    }
                    return { value: 0, literalValue };
                }
                case types_1.CodegenSchemaType.BOOLEAN: {
                    const literalValue = context.generator().toLiteral(false, options);
                    if (literalValue === null) {
                        return null;
                    }
                    return { value: false, literalValue };
                }
            }
            throw new Error(`Unsupported default value type: ${schemaType}`);
        }, initialValue: (options) => {
            const { required, schemaType, nativeType, defaultValue } = options;
            /*
              Default values in the spec are intended to be applied when a client or server receives a
              response or request, respectively, and values are missing.

              This implementation means that properties with defaults will get those default values as
              their initial value, meaning that any properties that are omitted in the _received_ request or
              response will have the default value.

              TODO But it also means that any requests or responses _sent_ will _also_ have the default values,
              rather than omitting the property and letting the receiving side apply the default value. This
              is NOT according to the spec and should be fixed.
             */
            if (defaultValue) {
                return defaultValue;
            }
            if (!required) {
                return null;
            }
            /*
              We create empty collections for required properties in the Java code. This is because we generate
              convenience methods for collections that initialise the collection when adding the first element,
              which would mean if we didn't initialise required collection properties we might end up sending
              a null collection value if the code didn't _add_ any elements. This would then require
              explicitly initialising each required collection in user code, either every time, or whenever
              no elements are added to it.
              
              Therfore we are not able to detect whether the user code has forgotten to populate the collection, like
              we are with scalar required properties, so we populate it with an empty collection so we always generate
              a valid result.
             */
            switch (schemaType) {
                case types_1.CodegenSchemaType.ARRAY:
                    /* Initialise required array properties with an empty array */
                    return { value: [], literalValue: `listOf<${nativeType.concreteType}>()` };
                case types_1.CodegenSchemaType.MAP:
                    /* Initialise empty map properties with an empty map */
                    return { value: {}, literalValue: `mapOf<${nativeType.concreteType}>()` };
                default:
                    return null;
            }
        }, operationGroupingStrategy: () => {
            return context.operationGroupingStrategies.addToGroupsByTagOrPath;
        }, allOfStrategy: () => types_1.CodegenAllOfStrategy.HIERARCHY, anyOfStrategy: () => types_1.CodegenAnyOfStrategy.OBJECT, oneOfStrategy: () => types_1.CodegenOneOfStrategy.NATIVE, supportsInheritance: () => false, supportsMultipleInheritance: () => false, nativeCompositionCanBeScope: () => false, nativeComposedSchemaRequiresName: () => true, nativeComposedSchemaRequiresObjectLikeOrWrapper: () => false, interfaceCanBeNested: () => true, watchPaths: () => {
            const result = [path_1.default.resolve(__dirname, '..', 'templates')];
            if (context.additionalWatchPaths) {
                result.push(...context.additionalWatchPaths());
            }
            if (generatorOptions.customTemplatesPath) {
                result.push(generatorOptions.customTemplatesPath);
            }
            return result;
        }, generatorType: () => types_1.CodegenGeneratorType.CLIENT, cleanPathPatterns: () => {
            const relativeSourceOutputPath = generatorOptions.relativeSourceOutputPath;
            const apiPackagePath = packageToPath(generatorOptions.apiPackage);
            const modelPackagePath = packageToPath(generatorOptions.modelPackage);
            const result = [
                path_1.default.join(relativeSourceOutputPath, apiPackagePath, '*Api.kt'),
                path_1.default.join(relativeSourceOutputPath, modelPackagePath, '*.kt'),
            ];
            if (context.additionalCleanPathPatterns) {
                result.push(...context.additionalCleanPathPatterns());
            }
            return result;
        }, templateRootContext: () => {
            return Object.assign(Object.assign(Object.assign({}, aCommonGenerator.templateRootContext()), generatorOptions), { generatorClass: '@openapi-generator-plus/kotlin-client-generator' });
        }, exportTemplates: async (outputPath, doc) => {
            const hbs = handlebars_1.default.create();
            (0, handlebars_templates_1.registerStandardHelpers)(hbs, context);
            hbs.registerHelper('stripLeadingSlash', function (text) {
                if (text.startsWith('/')) {
                    return text.substring(1, text.length);
                }
                else {
                    return text;
                }
            });
            await (0, handlebars_templates_1.loadTemplates)(path_1.default.resolve(__dirname, '..', 'templates'), hbs);
            if (context.loadAdditionalTemplates) {
                await context.loadAdditionalTemplates(hbs);
            }
            if (generatorOptions.customTemplatesPath) {
                await (0, handlebars_templates_1.loadTemplates)(generatorOptions.customTemplatesPath, hbs);
            }
            const rootContext = context.generator().templateRootContext();
            const relativeSourceOutputPath = generatorOptions.relativeSourceOutputPath;
            const apiPackagePath = packageToPath(generatorOptions.apiPackage);
            for (const group of doc.groups) {
                const operations = group.operations;
                if (!operations.length) {
                    continue;
                }
                await (0, handlebars_templates_1.emit)('api', path_1.default.join(outputPath, relativeSourceOutputPath, apiPackagePath, `${context.generator().toClassName(group.name)}Api.kt`), Object.assign(Object.assign(Object.assign({}, rootContext), group), { operations, servers: doc.servers }), true, hbs);
            }
            const modelPackagePath = packageToPath(generatorOptions.modelPackage);
            for (const schema of context.utils.values(doc.schemas)) {
                if ((0, types_1.isCodegenObjectSchema)(schema)) {
                    await (0, handlebars_templates_1.emit)('pojo', path_1.default.join(outputPath, relativeSourceOutputPath, modelPackagePath, `${context.generator().toClassName(schema.name)}.kt`), Object.assign(Object.assign({}, rootContext), { pojo: schema }), true, hbs);
                }
                else if ((0, types_1.isCodegenEnumSchema)(schema)) {
                    await (0, handlebars_templates_1.emit)('enum', path_1.default.join(outputPath, relativeSourceOutputPath, modelPackagePath, `${context.generator().toClassName(schema.name)}.kt`), Object.assign(Object.assign({}, rootContext), { enum: schema }), true, hbs);
                }
                else if ((0, types_1.isCodegenInterfaceSchema)(schema)) {
                    await (0, handlebars_templates_1.emit)('interface', path_1.default.join(outputPath, relativeSourceOutputPath, modelPackagePath, `${context.generator().toClassName(schema.name)}.kt`), Object.assign(Object.assign({}, rootContext), { interface: schema }), true, hbs);
                }
                else if ((0, types_1.isCodegenHierarchySchema)(schema)) {
                    await (0, handlebars_templates_1.emit)('hierarchy', path_1.default.join(outputPath, relativeSourceOutputPath, modelPackagePath, `${context.generator().toClassName(schema.name)}.kt`), Object.assign(Object.assign({}, rootContext), { hierarchy: schema }), true, hbs);
                }
                else if ((0, types_1.isCodegenOneOfSchema)(schema)) {
                    await (0, handlebars_templates_1.emit)('oneOf', path_1.default.join(outputPath, relativeSourceOutputPath, modelPackagePath, `${context.generator().toClassName(schema.name)}.kt`), Object.assign(Object.assign({}, rootContext), { oneOf: schema }), true, hbs);
                }
            }
            const gradle = generatorOptions.gradle;
            if (gradle) {
                await (0, handlebars_templates_1.emit)('build.gradle.kts', path_1.default.join(outputPath, 'build.gradle.kts'), Object.assign(Object.assign({}, rootContext), gradle), false, hbs);
            }
            /* Support */
            const supportPackagePath = packageToPath(generatorOptions.supportPackage);
            for (const file of await fs_1.promises.readdir(path_1.default.resolve(__dirname, '..', 'templates', 'support'))) {
                const fileBase = path_1.default.basename(file, path_1.default.extname(file));
                await (0, handlebars_templates_1.emit)(`support/${fileBase}`, path_1.default.join(outputPath, relativeSourceOutputPath, supportPackagePath, fileBase), Object.assign({}, rootContext), true, hbs);
            }
            /* Security */
            const securityPackagePath = packageToPath(generatorOptions.securityPackage);
            for (const file of await fs_1.promises.readdir(path_1.default.resolve(__dirname, '..', 'templates', 'security'))) {
                const fileBase = path_1.default.basename(file, path_1.default.extname(file));
                await (0, handlebars_templates_1.emit)(`security/${fileBase}`, path_1.default.join(outputPath, relativeSourceOutputPath, securityPackagePath, fileBase), Object.assign(Object.assign({}, rootContext), { securitySchemes: doc.securitySchemes }), true, hbs);
            }
            if (context.additionalExportTemplates) {
                await context.additionalExportTemplates(outputPath, doc, hbs, rootContext);
            }
        }, postProcessSchema(schema, helper) {
            function suggestedNameForType(type) {
                if (type.componentType) {
                    return `${type.componentType.nativeType}_array`;
                }
                else {
                    return type.nativeType;
                }
            }
            if ((0, types_1.isCodegenOneOfSchema)(schema)) {
                /* oneOf schemas turn into an enum, so each member needs a `name` for our enum */
                for (const comp of schema.composes) {
                    if (comp.name === null) {
                        comp.name = helper.uniqueName(suggestedNameForType(comp.nativeType), helper.scopeOf(schema), comp.schemaType);
                    }
                }
            }
        }, checkPropertyCompatibility: (parentProp, childProp) => {
            if (!baseGenerator.checkPropertyCompatibility(parentProp, childProp)) {
                return false;
            }
            /**
             * Because in Kotlin we use `Nullable` if a value of a property can be specifically "null",
             * properties are not compatible if their nullability varies.
             */
            if (!parentProp.nullable !== !childProp.nullable) {
                return false;
            }
            /**
             * Because in Kotlin we use an interface for allOf,
             * we can't comply if the required status mismatches at all,
             * for the same reason as nullability above.
             */
            if (parentProp.required !== childProp.required) {
                return false;
            }
            return true;
        } });
}
exports.default = createGenerator;
